from importlib import reload

import pytest

import aiodine
from aiodine import Store


@pytest.fixture(params=[Store, lambda: reload(aiodine)])
def store(request) -> Store:
    cls = request.param
    return cls()



import pytest
from aiodine import Store
from aiodine.exceptions import ProviderDoesNotExist

pytestmark = pytest.mark.asyncio


def declare_provider(store: Store):
    @store.provider
    async def setup():
        setup.called = True

    setup.called = False
    return setup


@pytest.mark.parametrize("lazy", (True, False))
async def test_from_string(store: Store, lazy: bool):
    if not lazy:
        setup = declare_provider(store)

    @store.consumer
    @store.useprovider("setup")
    async def consume():
        pass

    if lazy:
        setup = declare_provider(store)

    assert not setup.called
    await consume()
    assert setup.called


async def test_from_provider(store: Store):
    setup = declare_provider(store)

    @store.consumer
    @store.useprovider(setup)
    async def consume():
        pass

    assert not setup.called
    await consume()
    assert setup.called


async def test_use_unknown_provider(store: Store):
    @store.consumer
    @store.useprovider("foo")
    async def consume():
        pass

    with pytest.raises(ProviderDoesNotExist):
        await consume()



import pytest
from aiodine import Store

pytestmark = pytest.mark.asyncio


async def test_autouse_provider_is_injected_without_declaring_it(store: Store):
    used = False

    @store.provider(autouse=True)
    async def setup_stuff():
        nonlocal used
        used = True

    @store.consumer
    async def not_using_setup_stuff():
        pass

    await not_using_setup_stuff()
    assert used



import pytest
from inspect import iscoroutine

from aiodine import Store, scopes
from aiodine.exceptions import ProviderDeclarationError

pytestmark = pytest.mark.asyncio


async def test_use_async_provider(store: Store):
    @store.provider
    async def pitch():
        return "C#"

    @store.consumer
    def play_sync(pitch):
        return 2 * "C#"

    @store.consumer
    async def play_async(pitch):
        return 2 * "C#"

    assert await play_sync() == "C#C#"
    assert await play_async() == "C#C#"


async def test_lazy_async_provider(store: Store):
    @store.provider(lazy=True)
    async def pitch():
        return "C#"

    @store.consumer
    async def play(pitch):
        assert iscoroutine(pitch)
        return 2 * await pitch

    assert await play() == "C#C#"


async def test_lazy_provider_must_be_function_scoped(store: Store):
    with pytest.raises(ProviderDeclarationError):

        @store.provider(lazy=True, scope=scopes.SESSION)
        async def pitch():
            pass



import pytest

from aiodine import Store

pytestmark = pytest.mark.asyncio


async def test_override_provider_before_usage(store: Store):
    @store.provider
    async def hello():
        return "hello"

    @store.provider
    async def hello():
        return "HELLO"

    @store.consumer
    async def say(hello):
        return hello

    assert await say() == "HELLO"


async def test_override_provider_after_usage(store: Store):
    @store.provider
    async def hello():
        return "hello"

    @store.consumer
    async def say(hello):
        return hello

    @store.provider
    async def hello():
        return "HELLO"

    assert await say() == "HELLO"



import pytest

from aiodine import Store

pytestmark = pytest.mark.asyncio


async def test_sync_function_yield_provider(store: Store):
    setup = False
    teardown = False

    @store.provider
    def resource():
        nonlocal setup, teardown
        setup = True
        yield "resource"
        teardown = True

    @store.consumer
    def consume(resource: str):
        return resource.upper()

    assert await consume() == "RESOURCE"
    assert setup
    assert teardown


async def test_async_function_generator_provider(store: Store):
    setup = False
    teardown = False

    @store.provider
    async def resource():
        nonlocal setup, teardown
        setup = True
        yield "resource"
        teardown = True

    @store.consumer
    def consume(resource: str):
        return resource.upper()

    assert await consume() == "RESOURCE"
    assert setup
    assert teardown


async def test_session_generator_provider(store: Store):
    setup = False
    teardown = False

    @store.provider(scope="session")
    async def resource():
        nonlocal setup, teardown
        setup = True
        yield "resource"
        teardown = True

    @store.consumer
    def consume(resource: str):
        return resource.upper()

    assert await consume() == "RESOURCE"
    assert setup
    assert not teardown

    await store.exit_session()
    assert teardown


async def test_cleanup_even_if_exception_occurred(store: Store):
    teardown = False

    @store.provider
    async def resource():
        nonlocal teardown
        yield "resource"
        teardown = True

    @store.consumer
    async def consume(resource):
        raise ValueError

    with pytest.raises(ValueError):
        await consume()

    assert teardown



from asyncio import sleep, gather

import pytest
from aiodine import Store

# pylint: disable=no-value-for-parameter

pytestmark = pytest.mark.asyncio


async def test_provider_value_is_none_by_default(store: Store):
    store.create_context_provider("name")

    @store.consumer
    async def get_name(name):
        return name

    assert await get_name() is None


async def test_assign_value(store: Store):
    provider = store.create_context_provider("name")

    @store.consumer
    async def get_name(name):
        return name

    with provider.assign(name="alice"):
        assert await get_name() == "alice"

    assert await get_name() is None


async def test_multiple_providers(store: Store):
    provider = store.create_context_provider("name", "title")

    @store.consumer
    async def get_them(name, title):
        return name, title

    assert await get_them() == (None, None)

    with provider.assign(name="alice", title="Slim Fox"):
        assert await get_them() == ("alice", "Slim Fox")

    assert await get_them() == (None, None)


async def test_multi_client(store: Store):
    provider = store.create_context_provider("name")

    @store.consumer
    async def get_name(name):
        return name

    async def client1():
        with provider.assign(name="alice"):
            await sleep(0.01)
            # Would get "bob" if multiple clients were not supported.
            assert await get_name() == "alice"

    async def client2():
        await sleep(0.005)
        with provider.assign(name="bob"):
            await sleep(0.01)

    await gather(client1(), client2())



import inspect
import sys
from contextlib import suppress
from functools import WRAPPER_ASSIGNMENTS, partial, update_wrapper
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from .compat import AsyncExitStack, wrap_async
from .datatypes import CoroutineFunction
from .exceptions import ConsumerDeclarationError

if TYPE_CHECKING:  # pragma: no cover
    from .store import Store
    from .providers import Provider

PositionalProviders = List[Tuple[str, "Provider"]]
KeywordProviders = Dict[str, "Provider"]

# Sentinel for parameters that have no provider.
_NO_PROVIDER = object()


class ResolvedProviders(NamedTuple):

    positional: PositionalProviders
    keyword: KeywordProviders
    external: List["Provider"]


WRAPPER_IGNORE = {"__module__"}
if sys.version_info < (3, 7):  # pragma: no cover
    WRAPPER_IGNORE.add("__qualname__")

WRAPPER_ASSIGNMENTS = set(WRAPPER_ASSIGNMENTS) - WRAPPER_IGNORE
WRAPPER_SLOTS = {"__wrapped__", *WRAPPER_ASSIGNMENTS}


class Consumer:

    __slots__ = ("store", "func", "signature", *WRAPPER_SLOTS)

    def __init__(
        self,
        store: "Store",
        consumer_function: Union[partial, Callable, CoroutineFunction],
    ):
        self.store = store

        if isinstance(consumer_function, partial):
            if not inspect.iscoroutinefunction(consumer_function.func):
                raise ConsumerDeclarationError(
                    "'partial' consumer functions must wrap an async function"
                )
        else:
            if not inspect.isfunction(
                consumer_function
            ) and not inspect.ismethod(consumer_function):
                assert callable(consumer_function), "consumers must be callable"
                consumer_function = consumer_function.__call__

            if not inspect.iscoroutinefunction(consumer_function):
                consumer_function = wrap_async(consumer_function)

        self.func = consumer_function
        update_wrapper(
            self, self.func, assigned=WRAPPER_ASSIGNMENTS, updated=()
        )

    def resolve(self) -> ResolvedProviders:
        positional: PositionalProviders = []
        keyword: KeywordProviders = {}
        external = [
            *self.store.autouse_providers.values(),
            *self.store.get_used_providers(self.func),
        ]

        for name, parameter in inspect.signature(self.func).parameters.items():
            prov: Optional["Provider"] = self.store.providers.get(
                name, _NO_PROVIDER
            )

            if parameter.kind == inspect.Parameter.KEYWORD_ONLY:
                keyword[name] = prov
            else:
                positional.append((name, prov))

        return ResolvedProviders(
            positional=positional, keyword=keyword, external=external
        )

    async def __call__(self, *args, **kwargs):
        # TODO: cache providers after first call for better performance.
        providers = self.resolve()

        async with AsyncExitStack() as stack:

            async def _get_value(prov: "Provider"):
                if prov.lazy:
                    return prov(stack)
                return await prov(stack)

            for prov in providers.external:
                await _get_value(prov)

            # Create a stack out of the positional arguments.
            # Reverse it so we can `.pop()` out of it while
            # keeping the final order of arguments.
            args = list(reversed(args))

            injected_args = []
            for name, prov in providers.positional:
                if name in kwargs:
                    # Use values from keyword arguments in priority.
                    injected_args.append(kwargs.pop(name))
                    continue
                elif prov is _NO_PROVIDER:
                    # No provider exists. Use the next positional argument.
                    with suppress(IndexError):
                        injected_args.append(args.pop())
                else:
                    # A provider exists for this argument. Use it!
                    injected_args.append(await _get_value(prov))

            injected_kwargs = {}
            for name, prov in providers.keyword.items():
                if name in kwargs or prov is _NO_PROVIDER:
                    with suppress(KeyError):
                        injected_kwargs[name] = kwargs.pop(name)
                else:
                    injected_kwargs[name] = await _get_value(prov)

            return await self.func(*injected_args, **injected_kwargs)



class AiodineException(Exception):
    """Base exceptions for the aiodine package."""


class ConsumerDeclarationError(AiodineException):
    """Base exception for when a consumer is ill-declared."""


class ProviderDeclarationError(AiodineException):
    """Base exception for when a provider is ill-declared."""


class RecursiveProviderError(ProviderDeclarationError):
    """Raised when two providers depend on each other."""

    def __init__(self, first: str, second: str):
        message = (
            "recursive provider detected: "
            f"{first} and {second} depend on each other."
        )
        super().__init__(message)


class UnknownScope(AiodineException):
    """Raised when an unknown scope is used."""


class ProviderDoesNotExist(AiodineException):
    """Raised when using an unknown provider."""

    def __init__(self, name: str):
        super().__init__(f"provider {name} does not exist")



import pytest
from aiodine import Store

pytestmark = pytest.mark.asyncio


async def test_session_provider_instanciate_on_enter(store: Store):
    setup = False

    @store.provider(scope="session")
    async def resources():
        nonlocal setup
        setup = True
        return

    @store.consumer
    def consumer(resources: None):
        pass

    async with store.session():
        assert setup


async def test_session_provider_destroyed_on_exit(store: Store):
    @store.provider(scope="session")
    async def resources():
        return ["resource"]

    @store.consumer
    def consumer(resources: list):
        resources.append("other")
        return resources

    async with store.session():
        await consumer()
        assert await consumer() == ["resource", "other", "other"]

    # Instance was reset.
    assert await consumer() == ["resource", "other"]



import pytest

from aiodine import Store
from aiodine.exceptions import RecursiveProviderError


pytestmark = pytest.mark.asyncio


async def test_provider_uses_provider(store: Store):
    with store.exit_freeze():

        @store.provider
        def a():
            return "a"

        @store.provider
        def b(a):
            return a * 2

    func = store.consumer(lambda b: 2 * b)
    assert await func() == "aaaa"


async def test_provider_uses_provider_declared_later(store: Store):
    with store.exit_freeze():

        @store.provider
        def b(a):
            return a * 2

        @store.provider
        def a():
            return "a"

    func = store.consumer(lambda b: 2 * b)
    assert await func() == "aaaa"


async def test_detect_recursive_provider(store: Store):
    @store.provider
    def b(a):
        return a * 2

    with pytest.raises(RecursiveProviderError) as ctx:

        @store.provider
        def a(b):
            return a * 2



import sys

import pytest
from aiodine import Store


@pytest.fixture
def notes_module(store: Store):
    class NotesModule:
        @store.provider
        def pitch():  # pylint: disable=no-method-argument
            return "C#"

    sys.modules["notes"] = NotesModule


@pytest.mark.asyncio
@pytest.mark.usefixtures("notes_module")
async def test_discover_providers(store: Store):
    store.discover("notes")
    assert store
    assert store.has_provider("pitch")
    assert await store.consumer(lambda pitch: 2 * pitch)() == "C#C#"


def test_if_module_does_not_exist_then_error(store: Store):
    with pytest.raises(ImportError):
        store.discover("doesnotexist")



import inspect
from contextlib import contextmanager, suppress
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Union,
)

from . import scopes
from .compat import (
    AsyncExitStack,
    wrap_async,
    wrap_generator_async,
    ContextVar,
    Token,
)
from .datatypes import CoroutineFunction
from .exceptions import ProviderDeclarationError

if TYPE_CHECKING:  # pragma: no cover
    from .store import Store


async def _terminate_agen(async_gen: AsyncGenerator):
    with suppress(StopAsyncIteration):
        await async_gen.asend(None)


class Provider:
    """Base class for providers.

    This is mostly a wrapper around a provider function, along with
    some metadata.
    """

    __slots__ = ("func", "name", "scope", "lazy", "autouse")

    def __init__(
        self, func: Callable, name: str, scope: str, lazy: bool, autouse: bool
    ):
        if lazy and scope != scopes.FUNCTION:
            raise ProviderDeclarationError(
                "Lazy providers must be function-scoped"
            )

        if inspect.isgeneratorfunction(func):
            func = wrap_generator_async(func)
        elif inspect.isasyncgenfunction(func):
            pass
        elif not inspect.iscoroutinefunction(func):
            func = wrap_async(func)

        assert inspect.iscoroutinefunction(func) or inspect.isasyncgenfunction(
            func
        )

        self.func: Union[AsyncGenerator, CoroutineFunction] = func
        self.name = name
        self.scope = scope
        self.lazy = lazy
        self.autouse = autouse

    @classmethod
    def create(cls, func, **kwargs) -> "Provider":
        """Factory method to build a provider of the appropriate scope."""
        scope: Optional[str] = kwargs.get("scope")
        if scope == scopes.SESSION:
            return SessionProvider(func, **kwargs)
        return FunctionProvider(func, **kwargs)

    # NOTE: the returned value is an awaitable, so we *must not*
    # declare this function as `async` — its return value should already be.
    def __call__(self, stack: AsyncExitStack) -> Awaitable:
        raise NotImplementedError


class FunctionProvider(Provider):
    """Represents a function-scoped provider.

    Its value is recomputed every time the provider is called.
    """

    def __call__(self, stack: AsyncExitStack) -> Awaitable:
        value: Union[Awaitable, AsyncGenerator] = self.func()

        if inspect.isasyncgen(value):
            agen = value
            # We cannot use `await` in here => define a coroutine function
            # and return the (awaitable) coroutine itself.

            async def get_value() -> Any:
                # Executes setup + `yield <some_value>`.
                val = await agen.asend(None)
                # Registers cleanup to be executed when the stack exits.
                stack.push_async_callback(partial(_terminate_agen, agen))
                return val

            value: Awaitable = get_value()

        return value


class SessionProvider(Provider):
    """Represents a session-scoped provider.

    When called, it builds its instance if necessary and returns it. This
    means that the underlying provider is only built once and is reused
    across function calls.
    """

    __slots__ = Provider.__slots__ + ("_instance", "_generator")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._instance: Optional[Any] = None
        self._generator: Optional[AsyncGenerator] = None

    async def enter_session(self):
        value = self.func()

        if inspect.isawaitable(value):
            value = await value

        if inspect.isasyncgen(value):
            agen = value
            value = await agen.asend(None)
            self._generator = agen

        self._instance = value

    async def exit_session(self):
        if self._generator is not None:
            await _terminate_agen(self._generator)
            self._generator = None
        self._instance = None

    async def _get_instance(self) -> Any:
        if self._instance is None:
            await self.enter_session()
        return self._instance

    def __call__(self, stack: AsyncExitStack) -> Awaitable:
        return self._get_instance()


class ContextProvider:
    """A provider of context-local values.

    This provider is implemented using the ``contextvars`` module.

    Parameters
    ----------
    store : Store
    *names : str
        The name of the variables to provide. For each variable, a
        ``ContextVar`` is created and used by a new provider named
        after the variable.
    """

    def __init__(self, store: "Store", *names: str):
        self._store = store
        self._variables: Dict[str, ContextVar] = {}

        for name in names:
            self._build_provider(name)

    def _build_provider(self, name):
        self._variables[name] = ContextVar(name, default=None)

        async def provider():
            return self._variables[name].get()

        return self._store.provider(name=name)(provider)

    def _set(self, **values: Any) -> List[Token]:
        # Set new values for the given variables.
        tokens = []
        for name, val in values.items():
            token = self._variables[name].set(val)
            tokens.append(token)
        return tokens

    def _reset(self, *tokens: Token):
        # Reset variables to their previous value using the given tokens.
        for token in tokens:
            self._variables[token.var.name].reset(token)

    @contextmanager
    def assign(self, **values: Any):
        """Context manager to assign values to variables.

        Only the variables for the current context are changed. Values for
        other contexts are unaffected.

        Variables are reset to their previous value on exit.

        Parameters
        ----------
        **values : any
        """
        tokens = self._set(**values)
        try:
            yield
        finally:
            self._reset(*tokens)



import inspect
from functools import partial

import pytest
from aiodine import Store
from aiodine.exceptions import ConsumerDeclarationError

pytestmark = pytest.mark.asyncio


async def test_consumer_returns_coroutine_function_like(store: Store):
    func = store.consumer(lambda: "test")
    coro = func()
    assert inspect.isawaitable(coro)
    await coro


async def test_if_no_provider_decalred_then_behaves_like_func(store: Store):
    func = store.consumer(lambda: "test")
    assert await func() == "test"


async def test_if_provider_does_not_exist_then_missing_argument(store: Store):
    @store.provider
    def gra():
        return "gra"

    # "gra" exists, but not "arg"
    func = store.consumer(lambda arg: 2 * arg)

    with pytest.raises(TypeError):
        await func()

    assert await func(10) == 20


async def test_if_provider_exists_then_injected(store: Store):
    @store.provider
    def arg():
        return "foo"

    @store.consumer
    def func(arg):
        return 2 * arg

    assert await func() == "foofoo"


async def test_non_provider_parameters_after_provider_parameters_ok(
    store: Store
):
    @store.provider
    def pitch():
        return "C#"

    @store.consumer
    def play(pitch, duration):
        assert pitch == "C#"
        return (pitch, duration)

    assert await play(1) == ("C#", 1)
    assert await play(duration=1) == ("C#", 1)


async def test_non_provider_parameters_before_provider_parameters_ok(
    store: Store
):
    @store.provider
    def pitch():
        return "C#"

    @store.consumer
    def play(duration, pitch):
        assert pitch == "C#"
        return (pitch, duration)

    assert await play(1) == ("C#", 1)
    assert await play(duration=1) == ("C#", 1)


async def test_async_consumer(store: Store):
    @store.provider
    def pitch():
        return "C#"

    @store.consumer
    async def play(pitch):
        return 2 * pitch

    assert await play() == "C#C#"


@pytest.mark.parametrize("is_async", (True, False))
async def test_class_based_consumer(store: Store, is_async: bool):
    @store.provider
    def pitch():
        return "C#"

    class Consumer:
        if is_async:

            async def __call__(self, pitch: str):
                return 2 * pitch

        else:

            def __call__(self, pitch: str):
                return 2 * pitch

    consume = Consumer()
    consume = store.consumer(consume)
    assert await consume() == "C#C#"


async def test_method_consumer(store: Store):
    @store.provider
    async def pitch():
        return "C#"

    class Piano:
        async def play(self, pitch):
            return 2 * pitch

    piano = Piano()
    play = store.consumer(piano.play)
    assert await play() == "C#C#"


async def test_handle_keyword_only_parameters(store: Store):
    @store.provider
    async def pitch():
        return "C#"

    @store.consumer
    async def play(*, pitch, octave):
        return 2 * pitch + str(octave)

    assert await play(octave=2) == "C#C#2"
    assert await play(pitch="D", octave=2) == "DD2"


async def test_if_wrapper_then_wrapped_must_be_async(store: Store):
    @partial
    def not_async():
        pass

    with pytest.raises(ConsumerDeclarationError):
        store.consumer(not_async)


async def test_partial_of_async_function(store: Store):
    async def consume():
        return "OK"

    consumer = store.consumer(partial(consume))

    assert await consumer() == "OK"


async def test_keeps_order_of_arguments(store: Store):
    @store.consumer
    async def consume(a, b):
        return a + b

    assert await consume("a", "b") == "ab"



from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .store import Store


# NOTE: can't use @asynccontextmanager from contextlib because it was
# only added in Python 3.7.
class Session:
    def __init__(self, store: "Store"):
        self._store = store

    async def __aenter__(self):
        await self._store.enter_session()
        return None

    async def __aexit__(self, *args):
        await self._store.exit_session()



from aiodine import Store


def test_name_is_function_name_by_default(store: Store):
    @store.provider
    def example():
        pass

    assert example.name == "example"


def test_if_name_given_then_used(store: Store):
    @store.provider(name="an_example")
    def example():
        pass

    assert example.name == "an_example"



import pytest

from aiodine import Store

pytestmark = pytest.mark.asyncio


async def test_provider_factory_pattern(store: Store):
    @store.provider
    async def note():
        async def _get_note(pk: int):
            return {"id": pk}

        return _get_note

    @store.consumer
    async def get_note(pk: int, note):
        return await note(pk)

    assert await get_note(10) == {"id": 10}


async def test_generator_provider_factory(store: Store):
    setup = False
    cleanup = False

    @store.provider
    async def note():
        nonlocal setup, cleanup
        setup = True

        async def _get_note(pk: int):
            return {"id": pk}

        yield _get_note
        cleanup = True

    @store.consumer
    async def get_note(pk: int, note):
        return await note(pk)

    assert await get_note(10) == {"id": 10}
    assert setup
    assert cleanup



import inspect
from contextlib import contextmanager
from functools import partial
from importlib import import_module
from importlib.util import find_spec
from typing import Any, Callable, Dict, Optional, Union

from . import scopes
from .consumers import Consumer
from .datatypes import CoroutineFunction
from .exceptions import (
    RecursiveProviderError,
    UnknownScope,
    ProviderDoesNotExist,
)
from .providers import ContextProvider, Provider, SessionProvider
from .sessions import Session

DEFAULT_PROVIDER_MODULE = "providerconf"
_MISSING = object()


class Store:

    __slots__ = (
        "providers",
        "autouse_providers",
        "scope_aliases",
        "default_scope",
        "providers_module",
        "session_providers",
    )

    def __init__(
        self,
        providers_module=DEFAULT_PROVIDER_MODULE,
        scope_aliases: Dict[str, str] = None,
        default_scope: str = scopes.FUNCTION,
    ):
        if scope_aliases is None:
            scope_aliases = {}

        self.providers: Dict[str, Provider] = {}
        self.session_providers: Dict[str, SessionProvider] = {}
        self.autouse_providers: Dict[str, Provider] = {}
        self.scope_aliases = scope_aliases
        self.default_scope = default_scope
        self.providers_module = providers_module

    # Inspection.

    def empty(self):
        return not self.providers

    def has_provider(self, name: str) -> bool:
        return name in self.providers

    def _get(self, name: str, *, default: Any = _MISSING) -> Optional[Provider]:
        if default is not _MISSING:
            return self.providers.get(name, default)
        try:
            return self.providers[name]
        except KeyError:
            raise ProviderDoesNotExist(name) from None

    def _get_providers(self, func: Callable) -> Dict[str, Provider]:
        providers = {
            param: self._get(param, default=None)
            for param in inspect.signature(func).parameters
        }
        return dict(filter(lambda item: item[1] is not None, providers.items()))

    # Provider discovery.

    def discover_default(self):
        if find_spec(self.providers_module) is None:
            # Module does not exist.
            return
        self.discover(self.providers_module)

    @staticmethod
    def discover(*module_paths: str):
        for module_path in module_paths:
            import_module(module_path)

    # Provider registration.

    def provider(
        self,
        func: Callable = None,
        scope: str = None,
        name: str = None,
        lazy: bool = False,
        autouse: bool = False,
    ) -> Provider:
        if func is None:
            return partial(
                self.provider,
                scope=scope,
                name=name,
                lazy=lazy,
                autouse=autouse,
            )

        if scope is None:
            scope = self.default_scope
        else:
            scope = self.scope_aliases.get(scope, scope)

        if scope not in scopes.ALL:
            raise UnknownScope(scope)

        if name is None:
            name = func.__name__

        # NOTE: save the new provider before checking for recursion,
        # so that its dependants can detect it as a dependency.
        prov = Provider.create(
            func, name=name, scope=scope, lazy=lazy, autouse=autouse
        )
        self._add(prov)

        self._check_for_recursive_providers(name, func)

        return prov

    def _add(self, prov: Provider):
        self.providers[prov.name] = prov
        if isinstance(prov, SessionProvider):
            self.session_providers[prov.name] = prov
        if prov.autouse:
            self.autouse_providers[prov.name] = prov

    # Provider recursion check.

    def _check_for_recursive_providers(self, name: str, func: Callable):
        for other_name, other in self._get_providers(func).items():
            if name in self._get_providers(other.func):
                raise RecursiveProviderError(name, other_name)

    # Consumers.

    def consumer(
        self, consumer_function: Union[partial, Callable, CoroutineFunction]
    ) -> Consumer:
        return Consumer(self, consumer_function)

    # Used providers.

    def useprovider(self, *providers: Union[str, Provider]):
        def decorate(func):
            func.__useproviders__ = providers
            return func

        return decorate

    def get_used_providers(self, func: Callable):
        providers = getattr(func, "__useproviders__", [])
        return [
            prov if isinstance(prov, Provider) else self._get(prov)
            for prov in providers
        ]

    # Context providers.

    def create_context_provider(self, *args, **kwargs):
        return ContextProvider(self, *args, **kwargs)

    # Provider-in-providers freezing.

    def freeze(self):
        for prov in self.providers.values():
            prov.func = self.consumer(prov.func)

    @contextmanager
    def exit_freeze(self):
        yield
        self.freeze()

    # Sessions.

    async def enter_session(self):
        for provider in self.session_providers.values():
            await provider.enter_session()

    async def exit_session(self):
        for provider in self.session_providers.values():
            await provider.exit_session()

    def session(self):
        return Session(self)



import pytest

from aiodine import Store, scopes
from aiodine.exceptions import UnknownScope


@pytest.mark.parametrize(
    "store, expected_scope",
    [
        [Store(), scopes.FUNCTION],
        [Store(default_scope=scopes.FUNCTION), scopes.FUNCTION],
        [Store(default_scope=scopes.SESSION), scopes.SESSION],
    ],
)
def test_default_scope(store, expected_scope):
    @store.provider
    def items():
        pass

    assert items.scope == expected_scope


@pytest.mark.parametrize(
    "scope, expected_second_call",
    [(scopes.SESSION, [1, 2]), (scopes.FUNCTION, [2])],
)
@pytest.mark.asyncio
async def test_reuse_of_provided_values(
    store: Store, scope, expected_second_call
):
    @store.provider(scope=scope)
    def items():
        return []

    @store.consumer
    def add(items, value):
        items.append(value)
        return items

    assert await add(1) == [1]
    assert await add(2) == expected_second_call


@pytest.mark.parametrize(
    "aliases, scope, expected",
    [
        [
            {"foo": scopes.FUNCTION, "sass": scopes.SESSION},
            "foo",
            scopes.FUNCTION,
        ],
        [{"foo": scopes.FUNCTION}, "foo", scopes.FUNCTION],
        [{"sass": scopes.SESSION}, "sass", scopes.SESSION],
        [{"sass": scopes.SESSION}, scopes.FUNCTION, scopes.FUNCTION],
        [{"foo": scopes.FUNCTION}, scopes.SESSION, scopes.SESSION],
    ],
)
@pytest.mark.asyncio
async def test_scope_aliases(aliases, scope, expected):
    store = Store(scope_aliases=aliases)

    @store.provider(scope=scope)
    def items():
        return []

    assert items.scope == expected

    @store.consumer
    def add(items, value):
        items.append(value)
        return items

    assert await add(1) == [1]
    assert await add(2) == [1, 2] if items.scope == scopes.SESSION else [2]


def test_unknown_scope(store: Store):
    with pytest.raises(UnknownScope) as ctx:

        @store.provider(scope="blabla")
        def items():
            pass

    assert "blabla" in str(ctx.value)



from .providers import Provider
from .store import Store

# pylint: disable=invalid-name
_STORE = Store()

provider = _STORE.provider
consumer = _STORE.consumer
has_provider = _STORE.has_provider
useprovider = _STORE.useprovider
create_context_provider = _STORE.create_context_provider
providers_module = _STORE.providers_module
empty = _STORE.empty
discover = _STORE.discover
discover_default = _STORE.discover_default
freeze = _STORE.freeze
exit_freeze = _STORE.exit_freeze
session = _STORE.session
enter_session = _STORE.enter_session
exit_session = _STORE.exit_session

__version__ = "1.2.8"



import sys

from functools import wraps
from typing import AsyncGenerator, Awaitable, Callable, Generator

try:  # pragma: no cover
    from contextlib import AsyncExitStack  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    from async_exit_stack import AsyncExitStack


if sys.version_info < (3, 7):  # pragma: no cover
    from aiocontextvars import (  # pylint: disable=unused-import, import-error
        ContextVar,
        Token,
    )
else:  # pragma: no cover
    from contextvars import ContextVar, Token  # pylint: disable=unused-import


def wrap_async(func: Callable) -> Callable[..., Awaitable]:
    @wraps(func)
    async def async_func(*args, **kwargs):
        return func(*args, **kwargs)

    return async_func


def wrap_generator_async(gen: Generator) -> Callable[..., AsyncGenerator]:
    @wraps(gen)
    async def async_gen(*args, **kwargs):
        for item in gen(*args, **kwargs):
            yield item

    return async_gen



"""Package setup."""

import setuptools

description = "Async-first dependency injection library for Python"

with open("README.md", "r") as readme:
    long_description = readme.read()

GITHUB = "https://github.com/bocadilloproject/aiodine"

setuptools.setup(
    name="aiodine",
    version="1.2.8",
    author="Florimond Manca",
    author_email="florimond.manca@gmail.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "async_exit_stack;python_version<'3.7'",
        "aiocontextvars;python_version<'3.7'",
    ],
    python_requires=">=3.6",
    url=GITHUB,
    license="MIT",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Framework :: AsyncIO",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import sys

import pytest

from aiodine import Store


@pytest.fixture
def providers_module(store: Store):
    class FixtureConf:
        __spec__ = "spam"  # module-like

        @store.provider
        def example():  # pylint: disable=no-method-argument
            return "foo"

    sys.modules[store.providers_module] = FixtureConf


def test_default_providers_module(store: Store):
    assert store.providers_module == "providerconf"


def test_if_no_provider_module_then_ok(store: Store):
    store.discover_default()
    assert store.empty()


@pytest.mark.asyncio
@pytest.mark.usefixtures("providers_module")
async def test_if_provider_module_then_providers_are_loaded(store: Store):
    store.discover_default()
    assert not store.empty()
    assert store.has_provider("example")
    assert await store.consumer(lambda example: 2 * example)() == "foofoo"
